from twisted.internet.defer import succeed, inlineCallbacks
from yaml import dump

from ensemble.lib.mocker import MATCH
from ensemble.providers.dummy import MachineProvider
from ensemble.control import main

from .common import ControlToolTest


class ControlShutdownTest(ControlToolTest):

    @inlineCallbacks
    def test_shutdown_multiple_environments_no_default(self):
        """With multiple environments a default needs to be set or passed.
        """
        config = {
            "ensemble": "environments",
            "environments": {"firstenv": {"type": "dummy"},
                             "secondenv": {"type": "dummy"}}}
        self.write_config(dump(config))
        finished = self.setup_cli_reactor()
        self.setup_exit(1)
        self.mocker.replay()

        stderr = self.capture_stream('stderr')
        main(["shutdown"])
        yield finished
        msg = "There are multiple environments and no explicit default"
        self.assertIn(msg, stderr.getvalue())

    @inlineCallbacks
    def test_shutdown_invalid_environment(self):
        """If an invalid environment is specified, an error message is given.
        """
        config = {
            "ensemble": "environments",
            "environments": {"firstenv": {"type": "dummy"},
                             "secondenv": {"type": "dummy"}}}
        self.write_config(dump(config))
        finished = self.setup_cli_reactor()
        self.setup_exit(1)
        self.mocker.replay()

        stderr = self.capture_stream('stderr')
        main(["shutdown", "-e", "thirdenv"])
        yield finished
        msg = "Invalid environment 'thirdenv'"
        self.assertIn(msg, stderr.getvalue())

    @inlineCallbacks
    def test_shutdown_prompt_no(self):
        """If a user returns no to the prompt shutdown is aborted.
        """
        config = {
            "ensemble": "environments",
            "environments": {"firstenv": {"type": "dummy"},
                             "secondenv": {"type": "dummy"}}}
        self.write_config(dump(config))
        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        mock_raw = self.mocker.replace(raw_input)
        mock_raw(MATCH(lambda x: x.startswith("Warning, this will destroy")))
        self.mocker.result("n")
        self.mocker.replay()

        main(["shutdown", "-e", "secondenv"])
        yield finished
        msg = "Shutdown aborted"
        self.assertIn(msg, self.log.getvalue())

    @inlineCallbacks
    def test_shutdown(self):
        """
        'ensemble-control shutdown' will shutdown all configured instances
        in all environments.
        """
        config = {
            "ensemble": "environments",
            "environments": {"firstenv": {"type": "dummy"},
                             "secondenv": {"type": "dummy"}}}
        self.write_config(dump(config))
        finished = self.setup_cli_reactor()

        envs = set(("firstenv", "secondenv"))

        def track_shutdown_call(self):
            envs.remove(self.environment_name)
            return succeed(True)

        provider = self.mocker.patch(MachineProvider)
        provider.shutdown()
        self.mocker.call(track_shutdown_call, with_object=True)

        self.setup_exit(0)
        mock_raw = self.mocker.replace(raw_input)
        mock_raw(MATCH(lambda x: x.startswith("Warning, this will destroy")))
        self.mocker.result("y")
        self.mocker.replay()

        main(["shutdown", "-e", "firstenv"])
        yield finished
        msg = "Shutting down environment 'firstenv' (type: dummy)..."
        self.assertIn(msg, self.log.getvalue())
        self.assertEqual(envs, set(["secondenv"]))
