"""Implementation of formula-upgrade subcommand"""

import os

from twisted.internet.defer import inlineCallbacks

from ensemble.control.utils import get_environment

from ensemble.formula.errors import NewerFormulaNotFound
from ensemble.formula.metadata import parse_formula_id
from ensemble.formula.publisher import FormulaPublisher
from ensemble.formula.repository import LocalFormulaRepository

from ensemble.state.service import ServiceStateManager
from ensemble.unit.workflow import is_unit_running


def configure_subparser(subparsers):
    """Configure formula-upgrade subcommand"""
    sub_parser = subparsers.add_parser("upgrade-formula", help=command.__doc__)
    sub_parser.add_argument(
        "--dry-run", "-n", action="store_true",
        help="Dry-Run, show which formula would be deployed for upgrade.")
    sub_parser.add_argument(
        "--environment", "-e",
        help="Ensemble environment to operate in.")
    sub_parser.add_argument(
        "--repository", required=True,
        help="Directory for formula lookup and retrieval",
        type=os.path.abspath)
    sub_parser.add_argument(
        "service_name",
        help="Name of the service that should be upgraded")
    return sub_parser


def command(options):
    """Upgrade a service's formula."""
    environment = get_environment(options)
    return upgrade_formula(
        options.environments,
        environment,
        options.verbose,
        options.log,
        options.repository,
        options.service_name,
        options.dry_run)


@inlineCallbacks
def upgrade_formula(
    config, environment, verbose, log, repository_path, service_name, dry_run):
    """Upgrades a service's formula.

    First determines if an upgrade is available, then updates the
    service formula reference, and marks the units as needing upgrades.
    """
    provider = environment.get_machine_provider()
    client = yield provider.connect()

    repository = LocalFormulaRepository(repository_path)
    service_manager = ServiceStateManager(client)
    service_state = yield service_manager.get_service_state(service_name)
    formula_id = yield service_state.get_formula_id()

    namespace, name, current_revision = parse_formula_id(formula_id)

    # Lookup the formula in the repository
    formula = repository.find(name)

    # Verify its newer than what's deployed
    if not formula.metadata.revision > current_revision:
        if dry_run:
            log.info("Service already running latest formula %r", formula_id)
        else:
            raise NewerFormulaNotFound(formula_id)
    elif dry_run:
        log.info("Service would be upgraded from formula %r to %r",
                 formula_id, formula.metadata.get_formula_id("local"))

    # On dry run, stop before modifying state.
    if not dry_run:
        # Publish the new formula
        storage = provider.get_file_storage()
        publisher = FormulaPublisher(client, storage)
        yield publisher.add_formula("local", formula)
        result = yield publisher.publish()
        formula_state = result[0]

        # Update the service formula reference
        yield service_state.set_formula_id(formula_state.id)

    # Mark the units for upgrades
    units = yield service_state.get_all_unit_states()
    for unit in units:
        running, state = yield is_unit_running(client, unit)
        if not running:
            log.info(
                "Unit %r is not in a running state (state: %r), won't upgrade",
                unit.unit_name, state or "uninitialized")
            continue

        if not dry_run:
            yield unit.set_upgrade_flag()
