import hashlib
import tempfile
import os

from zipfile import ZipFile, BadZipfile

from ensemble.formula.base import FormulaBase
from ensemble.formula.config import ConfigOptions
from ensemble.formula.metadata import MetaData
from ensemble.errors import FormulaError
from ensemble.lib.filehash import compute_file_hash


class FormulaBundle(FormulaBase):
    """ZIP-archive that contains formula directory content."""

    def __init__(self, path):
        try:
            zf = ZipFile(path, 'r')
        except BadZipfile, exc:
            raise FormulaError(path, "Must be a ZIP-file (%s)." % str(exc))

        metadata = MetaData()
        if "metadata.yaml" not in zf.namelist():
            raise FormulaError(path,
                               "Archive does not contain required file: "
                               "\"metadata.yaml\".")

        content = zf.read("metadata.yaml")
        metadata.parse(content)
        self.metadata = metadata

        config = ConfigOptions()
        if "config.yaml" in zf.namelist():
            content = zf.read("config.yaml")
            config.parse(content)
        self.config = config

        self.path = isinstance(path, file) and path.name or path

    def compute_sha256(self):
        """Return the SHA256 digest for this formula bundle.

        The digest is extracted out of the final bundle file itself.
        """
        return compute_file_hash(hashlib.sha256, self.path)

    def extract_to(self, directory_path):
        """Extract the bundle to directory path and return a
        FormulaDirectory handle"""
        from .directory import FormulaDirectory
        zf = ZipFile(self.path, 'r')
        for info in zf.infolist():
            mode = info.external_attr >> 16
            extract_path = zf.extract(info, directory_path)
            os.chmod(extract_path, mode)
        return FormulaDirectory(directory_path)

    def as_bundle(self):
        return self

    def as_directory(self):
        """Returns the bundle as a FormulaDirectory using a temporary
        path"""
        dn = tempfile.mkdtemp(prefix="tmp")
        return self.extract_to(dn)
