from zookeeper import NodeExistsException, NoNodeException

from twisted.internet.defer import (
    DeferredList, inlineCallbacks, returnValue, succeed, FirstError)

from ensemble.state.formula import FormulaStateManager
from ensemble.state.errors import FormulaStateNotFound, StateChanged


class FormulaPublisher(object):
    """
    Publishes a formula to an environment.
    """

    def __init__(self, client, storage):
        self._client = client
        self._storage = storage
        self._formula_state_manager = FormulaStateManager(self._client)
        self._formula_add_queue = []
        self._formula_state_cache = {}

    @classmethod
    @inlineCallbacks
    def for_environment(cls, environment):
        provider = environment.get_machine_provider()
        storage = provider.get_file_storage()
        client = yield provider.connect()
        returnValue(cls(client, storage))

    @inlineCallbacks
    def add_formula(self, namespace, formula):
        """Schedule a formula for addition to an ensemble environment.

        Returns true if the formula is scheduled for upload, false if
        the formula is already present in ensemble.
        """
        formula_id = formula.metadata.get_formula_id(namespace)
        self._formula_add_queue.append((namespace, formula))
        if formula_id in self._formula_state_cache:
            returnValue(False)
        try:
            state = yield self._formula_state_manager.get_formula_state(
                formula_id)
        except FormulaStateNotFound:
            pass
        else:
            self._formula_state_cache[formula_id] = state
            returnValue(False)
        returnValue(True)

    def publish(self):
        """Publish all added formulas to provider storage and zookeeper.

        Returns the formula_state of all scheduled formulas.
        """
        publish_deferreds = []
        for namespace, formula in self._formula_add_queue:
            formula_id = formula.metadata.get_formula_id(namespace)

            if formula_id in self._formula_state_cache:
                publish_deferreds.append(
                    succeed(self._formula_state_cache[formula_id]))
                continue

            bundle = formula.as_bundle()

            formula_store_path = "formulas/%s:%s" % (
                formula_id,
                bundle.get_sha256())

            d = self._storage.put(formula_store_path, bundle.path)
            d.addCallback(self._cb_store_formula_state, namespace, bundle)
            d.addErrback(self._eb_verify_duplicate, namespace, bundle)
            publish_deferreds.append(d)

        publish_deferred = DeferredList(publish_deferreds,
                                        fireOnOneErrback=1,
                                        consumeErrors=1)
        # callbacks and deferreds to unwind the dlist
        publish_deferred.addCallback(self._cb_extract_formula_state)
        publish_deferred.addErrback(self._eb_extract_error)
        return publish_deferred

    def _cb_extract_formula_state(self, result):
        return [r[1] for r in result]

    def _eb_extract_error(self, failure):
        failure.trap(FirstError)
        return failure.value.subFailure

    def _cb_store_formula_state(self, result, namespace, formula):
        return self._formula_state_manager.add_formula_state(
            namespace, formula)

    @inlineCallbacks
    def _eb_verify_duplicate(self, failure, namespace, formula):
        """Detects duplicates vs. conflicts, raises stateerror on conflict."""
        failure.trap(NodeExistsException)
        formula_id = formula.metadata.get_formula_id(namespace)

        try:
            formula_state = \
                yield self._formula_state_manager.get_formula_state(formula_id)
        except NoNodeException:
            # Check if the state goes away due to concurrent removal
            msg = "Formula removed concurrently during publish, please retry."
            raise StateChanged(msg)

        if formula_state.get_sha256() != formula.get_sha256():
            msg = "Concurrent upload of formula has different checksum %s" % (
                formula_id)
            raise StateChanged(msg)
