import os
from twisted.python.failure import Failure
from ensemble.errors import EnsembleError, ProviderInteractionError

BOOTSTRAP_PACKAGES = [
    "bzr",
    "default-jre-headless",
    "zookeeper",
    "zookeeperd"]

DEFAULT_REPOSITORIES = [
    "ppa:ensemble/ppa"]

DEFAULT_PACKAGES = [
    "bzr",
    "tmux",
    "python-setuptools",
    "python-twisted",
    "python-argparse",
    "python-txaws",
    "python-zookeeper"]


def convert_unknown_error(failure):
    """
    Convert any non ensemble errors to a provider interaction error.

    Supports both usage from within an except clause, and as an
    errback handler ie. both the following forms are supported.

    ...
    try:
       something()
    except Exception, e:
       convert_unknown_errors(e)

    ...
    d.addErrback(convert_unknown_errors)
    """
    if isinstance(failure, Failure):
        error = failure.value
    else:
        error = failure

    if not isinstance(error, EnsembleError):
        error = ProviderInteractionError(error)

    if isinstance(failure, Failure):
        return Failure(error)
    raise error


def get_user_authorized_keys(config):
    """
    Locate a public key for the user. Either one explicitly passed
    in or one in the user's .ssh directory.

    @raise LookupError: Raised if a public ssh key is not found.
    """
    key_names = ["id_dsa.pub", "id_rsa.pub", "identity.pub"]
    if config.get("authorized-keys"):
        return config["authorized-keys"]

    if config.get("authorized-keys-path"):
        key_names[:] = [config.get("authorized-keys-path")]

    for key_path in key_names:
        path_candidate = os.path.expandvars(os.path.expanduser(key_path))
        if not os.path.exists(path_candidate):
            path_candidate = "~/.ssh/%s" % key_path
            path_candidate = os.path.expanduser(path_candidate)

        if not os.path.exists(path_candidate):
            continue
        return open(path_candidate).read()
    raise LookupError("SSH authorized/public key not found.")
