import os
import shutil
import tempfile

from txzookeeper import ZookeeperClient
from twisted.internet.defer import succeed, fail


from ensemble.errors import (
    FileNotFound, ProviderError, ProviderInteractionError)
from ensemble.machine import ProviderMachine


class DummyMachine(ProviderMachine):
    """Provider machine implementation specific to the dummy provider."""

    def __init__(self, instance_id, dns_name=None):
        super(DummyMachine, self).__init__(instance_id, dns_name)
        self._opened_ports = set()


class MachineProvider(object):

    def __init__(self, environment_name, config):
        self.environment_name = environment_name
        self.config = config
        self._machines = []
        self._state = None
        self._storage = None

    def connect(self, share=False):
        """Connect to the zookeeper ensemble running in the machine provider.

        @param share: Requests sharing of the connection with other clients
            attempting to connect to the same provider, if that's feasible.
            Unused for the dummy provider.
        """
        return ZookeeperClient(
            os.environ.get("ZOOKEEPER_ADDRESS", "127.0.0.1:2181"),
            session_timeout=1000).connect()

    def list_machines(self):
        """List all the machine running in the provider."""
        return succeed(list(self._machines))

    def start_machine(self, machine_data):
        """Start a machine in the provider."""
        if not "machine-id" in machine_data:
            return fail(ProviderError(
                "Machine state `machine-id` required in machine_data"))
        dns_name = machine_data.get("dns-name")
        machine = DummyMachine(len(self._machines), dns_name)
        self._machines.append(machine)
        return succeed([machine])

    def get_machine(self, provider_machine_id):
        """Retrieve a machine by provider machine id.
        """
        for machine in self._machines:
            if provider_machine_id == machine.instance_id:
                return succeed(machine)
        return fail(ProviderInteractionError("Machine (id: %r) was not found"))

    def bootstrap(self):
        """
        Bootstrap ensemble on the machine provider.
        """
        return self.start_machine({"machine-id": len(self._machines)})

    def shutdown(self):
        """
        Terminate any machine resources associated to the provider.
        """
        machines = self._machines
        self._machines = []
        return succeed(machines)

    def shutdown_machine(self, machine):
        """Terminate the given machine"""
        if not isinstance(machine, DummyMachine):
            return fail(ProviderError("Invalid machine for provider"))
        for m in self._machines:
            if m.instance_id == machine.instance_id:
                self._machines.remove(m)
                return succeed(m)
        return fail(ProviderError("Machine not found %r" % machine))

    def save_state(self, state):
        """Save the state to the provider."""
        self._state = state
        return succeed(None)

    def load_state(self):
        """Load the state from the provider."""
        if self._state:
            state = self._state
        else:
            state = {}
        return succeed(state)

    def get_file_storage(self):
        """Retrieve the C{FileStorage} provider abstracion."""
        if self._storage:
            return self._storage
        storage_path = self.config.get("storage-directory")
        if storage_path is None:
            storage_path = tempfile.mkdtemp()
        self._storage = FileStorage(storage_path)
        return self._storage

    def get_serialization_data(self):
        config = self.config.copy()
        # Introduce an additional variable to simulate actual
        # providers which may serialize additional values
        # from the environment or other external sources.
        config["dynamicduck"] = "magic"
        return config

    def open_port(self, machine, port, protocol="tcp"):
        """Dummy equivalent of ec2-authorize-group"""
        if not isinstance(machine, DummyMachine):
            return fail(ProviderError("Invalid machine for provider"))
        machine._opened_ports.add((port, protocol))
        return succeed(None)

    def close_port(self, machine, port, protocol="tcp"):
        """Dummy equivalent of ec2-revoke-group"""
        if not isinstance(machine, DummyMachine):
            return fail(ProviderError("Invalid machine for provider"))
        try:
            machine._opened_ports.remove((port, protocol))
        except KeyError:
            pass
        return succeed(None)

    def get_opened_ports(self, machine):
        """Dummy equivalent of ec2-describe-group

        This returns the current exposed ports in the environment for
        this machine. This directly goes against the provider. For
        EC2, this would be eventually consistent.
        """
        if not isinstance(machine, DummyMachine):
            return fail(ProviderError("Invalid machine for provider"))
        return succeed(machine._opened_ports)


class FileStorage(object):

    def __init__(self, path):
        self._path = path

    @property
    def path(self):
        # not part of the api, just for test cleanup
        return self._path

    def get(self, name):
        file_path = os.path.join(
            self._path, *filter(None, name.split("/")))
        if os.path.exists(file_path):
            return succeed(open(file_path))
        return fail(FileNotFound(file_path))

    def put(self, remote_path, local_path):
        local_path = os.path.abspath(local_path)
        if not os.path.exists(local_path):
            return fail(FileNotFound(local_path))
        store_path = os.path.join(
            self._path, *filter(None, remote_path.split("/")))
        store_path = os.path.abspath(store_path)
        if not store_path.startswith(self._path):
            return fail(AssertionError("Invalid Remote Path %s" % remote_path))

        parent_store_path = os.path.dirname(store_path)
        if not os.path.exists(parent_store_path):
            os.makedirs(parent_store_path)
        shutil.copyfile(local_path, store_path)
        return succeed(True)
