import time

from twisted.internet.defer import succeed, inlineCallbacks, returnValue

from txaws.ec2.exception import EC2Error

from ensemble.state.sshclient import SSHClient
from ensemble.errors import (
    NoConnection, EnvironmentNotFound, EnvironmentPending)
from operation import EC2Operation

from .utils import log


class EC2ZookeeperAware(object):

    def _verify_zookeeper_instances(self, state):
        """Verify which zookeeper nodes are running.

        Returns the first valid instance found as a single element list,
        or returns an empty list. If
        """
        instances = None

        if state:
            instances = state.get("zookeeper-instances", None)

        if not instances:
            d = succeed([])
        else:
            instance_id = instances.pop(0)
            d = self.ec2.describe_instances(instance_id)

        d.addCallback(self._cb_filter_running_zookeeper_instances)
        d.addErrback(self._eb_invalid_instance_id, instances)

        return d

    def _eb_invalid_instance_id(self, failure, instances):
        # Trap these failures to continue processing instances
        failure.trap(
            EC2Error, EnvironmentPending, EnvironmentNotFound)

        # Reformat invalid ids ec2 errors into an appros ensemble error.
        if not instances and failure.check(EC2Error):
            error_code = failure.value.get_error_codes()
            if error_code == "InvalidInstanceID.NotFound":
                raise EnvironmentNotFound(
                    "No Ensemble machines found. "
                    "Is the environment bootstrapped?")

        # Process next instance
        if instances:
            instance_id = instances.pop(0)
            d = self.ec2.describe_instances(instance_id)

            d.addCallback(self._cb_filter_running_zookeeper_instances)
            d.addErrback(self._eb_invalid_instance_id, instances)
            return d

        # Return the original failure
        return failure

    def _cb_filter_running_zookeeper_instances(self, instances):
        # instances is either a single element list or an empty list.
        assert len(instances) <= 1, "Should only process single instance %r"
        for instance in instances:
            if instance.instance_state == "running":
                return [instance]
            elif instance.instance_state == "pending":
                raise EnvironmentPending(
                    "Started machine is still pending.")
        raise EnvironmentNotFound(
            "No Ensemble machines found. "
            "Is the environment bootstrapped?")


class EC2Connect(EC2Operation, EC2ZookeeperAware):

    recent_machine_delay = 5 * 60

    def run(self, share=False):
        state_deferred = self.provider.load_state()
        state_deferred.addCallback(self._verify_zookeeper_instances)
        state_deferred.addCallback(self._connect_to_instance, share)
        state_deferred.addCallback(self._wait_for_initialization)
        return state_deferred

    def _connect_to_instance(self, instances, share):
        # TODO Should we pick a random entry from the nodes list?
        log.info("Connecting to environment.")
        client = SSHClient()
        instance = instances[0]
        result = client.connect(instance.dns_name + ":2181",
                                timeout=30, share=share)
        result.addErrback(self._check_machine_age, instance)
        return result

    def _check_machine_age(self, failure, instance):
        failure.trap(NoConnection)
        if instance.launch_time > time.time() - self.recent_machine_delay:
            raise NoConnection("Can't yet connect to started machine: %s" %
                               str(failure.value))
        return failure

    @inlineCallbacks
    def _wait_for_initialization(self, client):
        exists_d, watch_d = client.exists_and_watch("/initialized")
        exists = yield exists_d
        if not exists:
            log.info("Environment still initializing.  Will wait.")
            yield watch_d
        else:
            log.debug("Environment already initialized.")
        returnValue(client)
