from twisted.internet.defer import inlineCallbacks, returnValue, fail
from ensemble.providers import common
from ensemble.state.auth import make_identity
from ensemble.errors import EnvironmentNotFound, ProviderError

from .machine import EC2ProviderMachine
from .connect import EC2ZookeeperAware
from .operation import EC2Operation
from .utils import log, get_machine_options


DEFAULT_AVAILABILITY_ZONE = "us-east-1b"

ENSEMBLE_UPDATE_TEMPLATE = \
    "cd /usr/lib/ensemble/ensemble && sudo /usr/bin/bzr %s"

ENSEMBLE_UPDATE = ENSEMBLE_UPDATE_TEMPLATE % "up"


class EC2LaunchMachine(EC2Operation, EC2ZookeeperAware):

    def run(self, machine_data):
        """Launch a machine in the EC2 provider.

        @param machine_data keyowrd arguments passed to the machine
               via as a serialized yaml data structure. Must include
               a `machine-id` key, denoting the new machine's machine
               state id.

        Returns a C{EC2ProviderMachine} instance.
        """
        if not "machine-id" in machine_data:
            return fail(ProviderError(
                "Machine state `machine-id` not provided in machine_data."))
        launch_deferred = self.launch(machine_data=machine_data)
        launch_deferred.addCallback(self._on_launched)
        return launch_deferred

    def _on_launched(self, instance_ids):
        return [EC2ProviderMachine(id) for id in instance_ids]

    def _ensure_group(self, groups):
        """Ensure the ensemble group is the machine launch groups.

        Machines launched by ensemble are tagged with a group so they can
        be distinguished from other machines that might be running on
        an EC2 account. This group can be specified explicitly or implicitly
        defined by the environment name.

        @param security_groups: The configured security groups for a machine
        instance.
        """
        # Label the instance with the ensemble name.
        d = self.ec2.describe_security_groups()
        d.addCallback(self._on_received_groups, groups)
        return d

    def _on_received_groups(self, security_groups, groups):
        ensemble_group = "ensemble-%s" % self.provider.environment_name
        group_ids = [group.name for group in security_groups]

        # Ensure the provider group is added to the instance groups.
        if not ensemble_group in groups:
            groups.append(ensemble_group)

        # Create the provider group if doesn't exist.
        if not ensemble_group in group_ids:
            log.debug("Creating ensemble provider group %s", ensemble_group)
            group_created = self.ec2.create_security_group(
                ensemble_group,
                "Ensemble group for %s" % self.provider.environment_name)

            # Authorize SSH.
            group_created.addCallback(
                lambda x: self.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="tcp",
                    from_port="22", to_port="22",
                    cidr_ip="0.0.0.0/0"))

            # We need to describe the group to pickup the owner_id for auth.
            group_created.addCallback(
                lambda x: self.ec2.describe_security_groups(ensemble_group))

            # Authorize Internal ZK Traffic
            group_created.addCallback(
                lambda groups: self.ec2.authorize_security_group(
                    ensemble_group,
                    source_group_name=ensemble_group,
                    source_group_owner_id=groups.pop().owner_id))

            # TODO: For now authorize external traffic to the instance,
            # except for the ZooKeeper port.  The expose functionality
            # that is in progress will address this in a better way.
            group_created.addCallback(
                lambda groups: self.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="tcp",
                    from_port="1",
                    to_port="2180",
                    cidr_ip="0.0.0.0/0"))
            group_created.addCallback(
                lambda groups: self.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="tcp",
                    from_port="2182",
                    to_port="65535",
                    cidr_ip="0.0.0.0/0"))

            group_created.addCallback(
                lambda groups: self.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="udp",
                    from_port="1",
                    to_port="2180",
                    cidr_ip="0.0.0.0/0"))
            group_created.addCallback(
                lambda groups: self.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="udp",
                    from_port="2182",
                    to_port="65535",
                    cidr_ip="0.0.0.0/0"))

            # We use a separate callback here, because we need to wait
            # on the authorize security group deferred to succeed before
            # we can return the groups result.
            return group_created.addCallback(lambda x: groups)
        return groups

    def get_launch_scripts(self, machine_data):
        """Get the scripts that will execute after the machine has launched."""

        if machine_data.get("ensemble-branch"):
            ensemble_vcs = ENSEMBLE_UPDATE_TEMPLATE % (
                "switch %s" % machine_data["ensemble-branch"])
        else:
            ensemble_vcs = ENSEMBLE_UPDATE

        launch_scripts = [
            "sudo mkdir -p /usr/lib/ensemble",
            ensemble_vcs,
            "cd /usr/lib/ensemble/ensemble && sudo python setup.py develop",
            "cd /usr/lib/ensemble/txzookeeper && sudo /usr/bin/bzr up",
            "cd /usr/lib/ensemble/txzookeeper && sudo python setup.py develop"]

        machine_agent_script = (
            "ENSEMBLE_MACHINE_ID=%(machine-id)s "
            "ENSEMBLE_ZOOKEEPER=%(ensemble-zookeeper-hosts)s "
            "python -m ensemble.agents.machine -n "
            "--logfile=/var/log/ensemble/machine-agent.log "
            "--pidfile=/var/run/ensemble/machine-agent.pid")
        launch_scripts.append(machine_agent_script % machine_data)

        return launch_scripts

    def get_machine_variables(self, machine_data):
        """Get the variables used for launching a machine and its cloud-init"""
        # If we're using a custom branch, pass it via cloud-init to the agents,
        # so the entire cluster uses it.
        ensemble_branch = self.config.get("ensemble-branch")
        if ensemble_branch:
            machine_data["ensemble-branch"] = ensemble_branch

        scripts = self.get_launch_scripts(machine_data)
        variables = {
            "ensemble-provider-type": "ec2",
            "packages": list(common.DEFAULT_PACKAGES),
            "repositories": list(common.DEFAULT_REPOSITORIES),
            "scripts": scripts}

        machine_data.update(variables)
        return machine_data

    @inlineCallbacks
    def launch(self, machine_data, send_zoo_hosts=True):
        """Launch an instance node within the machine provider environment.

        @param machine_data a dictionary of data that is passed along to
               provided to the machine as serialized yaml. 'machine-id' is a
               required key, denoting the machine's zookeeper node for its
               machine agent.
        @param send_zoo_hosts boolean parameter, if true machines are passed
               the zookeeper hosts as part of their cloud-init configuration.
        """

        # Get the machine launch config/options
        authorized_keys = common.get_user_authorized_keys(self.config)

        # Get the ensemble zookeeper hosts for the node's agents to connect to.
        if send_zoo_hosts:
            zookeeper_hosts = yield self.get_ensemble_zookeepers()
            # format multiple hosts in string suitable for client
            hosts = ", ".join(["%s:2181" % h for h in zookeeper_hosts])
            machine_data["ensemble-zookeeper-hosts"] = hosts

        variables = self.get_machine_variables(machine_data)

        # Retrieves standard ec2 run_instances arguments, of note
        # image id, and instance type.
        machine_options = yield get_machine_options(
            self.config,
            authorized_keys=authorized_keys,
            **variables)

        # Ensure the ensemble security group exists and is included in the
        # machine instance launch options.
        machine_options["security_groups"] = yield self._ensure_group(
            machine_options["security_groups"])

        # Launch the instance.
        instances = yield self.ec2.run_instances(**machine_options)

        returnValue(
            [instance.instance_id for instance in instances])

    @inlineCallbacks
    def get_ensemble_zookeepers(self):
        state = yield self.provider.load_state()
        zookeeper_instances = yield self._verify_zookeeper_instances(state)
        hosts = [i.private_dns_name for i in zookeeper_instances]
        returnValue(hosts)


class EC2Bootstrap(EC2LaunchMachine):
    """Amazon EC2 C{MachineProvider} environment bootstrap."""

    def run(self):
        """Launch a bootstrap machine on ec2.

        Run any bootstrap required to get a working ensemble within the machine
        provider. Returns the corresponding ensemble instances' public dns
        names within the account.
        """
        load_deferred = self.provider.load_state()
        load_deferred.addCallback(self._verify_zookeeper_instances)
        load_deferred.addCallback(self._on_existing_state_found)
        load_deferred.addErrback(self._on_no_state_found)
        return load_deferred

    def _on_existing_state_found(self, state):
        log.info("Ensemble environment previously bootstrapped.")
        return [EC2ProviderMachine(instance.instance_id, instance.dns_name)
                for instance in state]

    def _on_no_state_found(self, failure):
        failure.trap(EnvironmentNotFound)

        log.debug("Launching Ensemble bootstrap instance.")

        # Disable searching for ensemble hosts, there are none.
        launch_deferred = self.launch(
            send_zoo_hosts=False,
            machine_data={"machine-id": "0"})

        # Save the state of the ensemble instances to S3.
        launch_deferred.addCallback(self._on_bootstrap_launched)
        launch_deferred.addCallback(self._on_launched)
        return launch_deferred

    def _on_bootstrap_launched(self, instances):
        d = self.provider.save_state({"zookeeper-instances": instances})
        d.addCallback(lambda x: instances)
        return d

    def get_machine_variables(self, machine_data):
        """Get the variables used for launching a machine and its cloud-init"""
        machine_data["ensemble-zookeeper-hosts"] = "localhost:2181"
        variables = super(EC2Bootstrap, self).get_machine_variables(
            machine_data)
        variables["packages"].extend(common.BOOTSTRAP_PACKAGES)
        admin_identity = make_identity(
            "admin:%s" % self.config["admin-secret"])

        initialize_script = (
            'sudo /usr/local/bin/ensemble-admin initialize --instance-id=%s'
            ' --admin-identity="%s"') % (
            "$(curl http://169.254.169.254/1.0/meta-data/instance-id)",
            admin_identity)

        # Initialize needs to come before machine agent.
        variables["scripts"].insert(-1, initialize_script)

        provision_agent_script = (
            "ENSEMBLE_ZOOKEEPER=%(ensemble-zookeeper-hosts)s "
            "python -m ensemble.agents.provision -n "
            "--logfile=/var/log/ensemble/provision-agent.log "
            "--pidfile=/var/run/ensemble/provision-agent.pid")

        variables["scripts"].append(provision_agent_script % machine_data)

        return variables
