import logging

from yaml import dump, load

from twisted.internet.defer import succeed
from txaws.ec2.model import Instance, SecurityGroup

from ensemble.errors import (
    EnvironmentNotFound, EnvironmentPending, ProviderError)
from ensemble.providers.ec2.machine import EC2ProviderMachine
from ensemble.providers.ec2 import MachineProvider
from ensemble.providers.ec2.launch import EC2Bootstrap, EC2LaunchMachine
from ensemble.providers.common import (
    DEFAULT_REPOSITORIES, DEFAULT_PACKAGES, BOOTSTRAP_PACKAGES)
from ensemble.state.auth import make_identity
from ensemble.tests.common import get_test_zookeeper_address

from ensemble.lib.testing import TestCase
from ensemble.lib.mocker import MATCH

from .common import EC2TestMixin


MATCH_AMI = MATCH(lambda image_id: image_id.startswith("ami-"))


class MachineLaunchMixin(object):

    default_zookeeper_host = Instance(
        "i-es-zoo", "running", private_dns_name="es.example.internal")

    def _mock_launch_utils(self):
        get_public_key = self.mocker.replace(
            "ensemble.providers.common.get_user_authorized_keys")

        def match_config(arg):
            return isinstance(arg, dict)

        get_public_key(MATCH(match_config))
        self.mocker.result("zebra")
        # AMI is not retrieve dynamically at the moment, default values are
        # utilized.
        #get_ami = self.mocker.replace(
        #    "ensemble.providers.ec2.utils.get_current_ami")
        #get_ami()
        #self.mocker.result(succeed("ami-714ba518"))

    def _mock_create_group(self):
        group_name = "ensemble-%s" % self.env_name
        self.ec2.create_security_group(
            group_name, "Ensemble group for %s" % self.env_name)
        self.mocker.result(succeed(True))

        self.ec2.authorize_security_group(
            group_name, ip_protocol="tcp", from_port="22",
            to_port="22", cidr_ip="0.0.0.0/0")
        self.mocker.result(succeed([self.env_name]))

        self.ec2.describe_security_groups(group_name)
        self.mocker.result(succeed(
            [SecurityGroup(group_name, "", owner_id="123")]))

        self.ec2.authorize_security_group(
            group_name, source_group_name=group_name,
            source_group_owner_id="123")
        self.mocker.result(succeed(True))

        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="tcp",
            from_port="1",
            to_port="2180",
            cidr_ip="0.0.0.0/0")
        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="tcp",
            from_port="2182",
            to_port="65535",
            cidr_ip="0.0.0.0/0")

        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="udp",
            from_port="1",
            to_port="2180",
            cidr_ip="0.0.0.0/0")
        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="udp",
            from_port="2182",
            to_port="65535",
            cidr_ip="0.0.0.0/0")

        self.mocker.result(succeed(True))

    def _mock_get_zookeeper_hosts(self, hosts=None):
        """
        Try to encapsulate a variety of behaviors here..

        if hosts is None, a default host is used.
        if hosts is False, no s3 state is returned
        if hosts are passed as a list of instances, they
           are returned.
        """

        if hosts is None:
            hosts = [self.default_zookeeper_host]

        self.s3.get_object(self.env_name, "provider-state")
        if hosts is False:
            self.mocker.result(succeed(None))
            return

        state = dump({
            "zookeeper-instances":
            [i.instance_id for i in hosts]})

        self.mocker.result(succeed(state))
        if hosts:
            # connect grabs the first host of a set.
            self.ec2.describe_instances(hosts[0].instance_id)
            self.mocker.result(succeed([hosts[0]]))


class EC2MachineLaunchTest(EC2TestMixin, MachineLaunchMixin, TestCase):

    def _mock_launch(self, instance=None, custom_verify=None):

        def verify_user_data(data):
            lines = data.split("\n")
            self.assertEqual(lines.pop(0), "#cloud-config")
            config = load("\n".join(lines))
            repos = [dict(source=r) for r in DEFAULT_REPOSITORIES]
            self.assertEqual(config["apt_sources"], repos)
            self.assertEqual(config["packages"], DEFAULT_PACKAGES)
            self.assertEqual(config["machine-data"]["machine-id"], "machine-1")

            if custom_verify:
                custom_verify(config)
            return True

        self.ec2.run_instances(
            image_id=MATCH_AMI,
            instance_type="m1.small",
            max_count=1,
            min_count=1,
            security_groups=["%s-%s" % ("ensemble", self.env_name)],
            user_data=MATCH(verify_user_data))

        if instance:
            self.mocker.result(succeed([instance]))
        else:
            self.mocker.result(succeed([]))

    def test_provider_launch(self):
        """
        The provider can be used to launch a machine with a minimal set of
        required packages, repositories, and and security groups.
        """
        instance = Instance("i-foobar", "running", dns_name="x1.example.com")

        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_get_zookeeper_hosts()
        self._mock_launch(instance)
        self.mocker.replay()

        def verify_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, EC2ProviderMachine))
            self.assertEqual(result.instance_id, instance.instance_id)

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        d.addCallback(verify_result)
        return d

    def test_machine_id_required_or_provider_error_raised(self):
        self.mocker.replay()
        provider = self.get_provider()
        d = provider.start_machine({})
        return self.assertFailure(d, ProviderError)

    def test_ensemble_branch_config(self):
        """A custom ensemble branch can be specified in the provider config.

        Which will be utilized when launching machines.
        """
        self.mocker.replay()
        config = self.get_config()
        custom_branch = "lp:~nobody/ensemble/cool-branch"
        config["ensemble-branch"] = custom_branch

        provider = MachineProvider(self.env_name, config)
        launch = EC2LaunchMachine(provider)
        variables = launch.get_machine_variables(
            machine_data={
                "machine-id": "machine-2",
                "ensemble-zookeeper-hosts": get_test_zookeeper_address()})
        cmd = "cd /usr/lib/ensemble/ensemble && sudo /usr/bin/bzr switch %s"%(
            custom_branch)
        self.failUnlessIn(cmd, variables["scripts"])
        self.assertEqual(variables["ensemble-branch"], custom_branch)

    def test_zookeeper_hosts_provided(self):
        """
        The internal ip address of Ensemble zookeeper hosts are passed
        to launched machines.
        """
        instances = [
            Instance(
                "i-foobar", "running",
                private_dns_name="x1.internal",
                dns_name="x1.example.com"),
            Instance(
                "i-foobar-2", "running",
                private_dns_name="x2.internal",
                dns_name="x2.example.com")]

        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_get_zookeeper_hosts(instances)

        def verify_host(data):
            host = instances[0]
            self.assertEqual(
                data["machine-data"]["ensemble-zookeeper-hosts"],
                "%s:2181" % host.private_dns_name)

        self._mock_launch(instances[1], verify_host)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        return d

    def test_launch_with_no_ensemble_s3_state(self):
        """
        Attempting to launch without any ensemble saved state, means
        we can't provide a way for a launched instance to connect
        to the zookeeper shared state. An assertion error is raised
        instead of allowing this.
        """
        self._mock_get_zookeeper_hosts(False)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        self.assertFailure(d, EnvironmentNotFound)
        return d

    def test_launch_with_no_ensemble_zookeeper_hosts(self):
        """
        Attempting to launch without any ensemble zookeeper hosts, means
        we can't provide a way for a launched instance to connect
        to the zookeeper shared state. An assertion error is raised
        instead of allowing this.
        """
        self._mock_get_zookeeper_hosts([])
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        self.assertFailure(d, EnvironmentNotFound)
        return d

    def test_launch_starts_machine_agent(self):
        """A launch'd machine should have a machine agent."""
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_get_zookeeper_hosts()

        def verify_machine_agent(data):
            script = ("ENSEMBLE_MACHINE_ID=machine-1 "
                      "ENSEMBLE_ZOOKEEPER=es.example.internal:2181 "
                      "python -m ensemble.agents.machine "
                      "-n --logfile=/var/log/ensemble/machine-agent.log "
                      "--pidfile=/var/run/ensemble/machine-agent.pid")
            self.assertEqual(script, data["runcmd"][-1])

        self._mock_launch(custom_verify=verify_machine_agent)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        return d

    def test_provider_type_machine_variable(self):
        """The provider type is available via cloud-init."""
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_get_zookeeper_hosts()

        def verify_provider_type(data):
            self.assertEqual(
                data["machine-data"]["ensemble-provider-type"], "ec2")

        self._mock_launch(custom_verify=verify_provider_type)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        return d


class EC2BootstrapTest(EC2TestMixin, MachineLaunchMixin, TestCase):

    def _mock_save(self):
        """Mock saving bootstrap instances to S3."""
        self.s3.get_bucket(self.env_name)
        self.mocker.result(succeed(True))

        def match_string(data):
            return isinstance(data, str)

        self.s3.put_object(
            self.env_name, "provider-state",
            MATCH(match_string),
            content_type="text/plain")
        self.mocker.result(succeed(True))

    def _mock_launch(self):
        """Mock launching a bootstrap machine on ec2."""
        credentials = "admin:%s" % self.get_config()["admin-secret"]
        admin_identity = make_identity(credentials)

        def verify_user_data(data):

            lines = data.split("\n")
            self.assertEqual(lines.pop(0), "#cloud-config")
            config = load("\n".join(lines))
            repos = [dict(source=r) for r in DEFAULT_REPOSITORIES]
            self.assertEqual(config["apt_sources"], repos)
            self.assertEqual(
                config["packages"],
                list(DEFAULT_PACKAGES) + BOOTSTRAP_PACKAGES)
            self.failUnlessIn("admin-identity", config["runcmd"][-3])

            script = (
                'sudo /usr/local/bin/ensemble-admin'
                " initialize --instance-id=%s"
                ' --admin-identity="%s"') % (
                    "$(curl http://169.254.169.254/1.0/meta-data/instance-id)",
                    admin_identity)

            self.assertEqual(config["runcmd"][-3], script)

            script = (
                "ENSEMBLE_MACHINE_ID=0 ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.machine -n "
                "--logfile=/var/log/ensemble/machine-agent.log "
                "--pidfile=/var/run/ensemble/machine-agent.pid")

            self.assertEqual(config["runcmd"][-2], script)

            provision_agent_script = (
                "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.provision -n "
                "--logfile=/var/log/ensemble/provision-agent.log "
                "--pidfile=/var/run/ensemble/provision-agent.pid")
            self.assertEqual(config["runcmd"][-1], provision_agent_script)
            return True

        self.ec2.run_instances(
            image_id=MATCH_AMI,
            instance_type="m1.small",
            max_count=1,
            min_count=1,
            security_groups=["ensemble-%s" % self.env_name],
            user_data=MATCH(verify_user_data))

    def test_launch_bootstrap(self):
        """The provider bootstrap can launch a bootstrap/zookeeper machine."""

        log = self.capture_logging("ensemble.ec2", level=logging.DEBUG)

        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed([]))
        self._mock_save()
        self.mocker.replay()

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        deferred = bootstrap.run()

        def check_log(result):
            log_text = log.getvalue()
            self.assertIn("Launching Ensemble bootstrap instance", log_text)
            self.assertNotIn("previously bootstrapped", log_text)

        deferred.addCallback(check_log)
        return deferred

    def test_launch_bootstrap_existing_provider_group(self):
        """
        When launching a bootstrap instance the provider will use an existing
        provider instance group.
        """
        self.capture_logging("ensemble.ec2")
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed([]))
        self._mock_save()
        self.mocker.replay()

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        return bootstrap.run()

    def test_run_with_loaded_state(self):
        """
        If the provider bootstrap is run when there is already a running
        bootstrap instance, it will just return the existing instance.
        """
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]
        state = dump({
            "zookeeper-instances":
            [i.instance_id for i in instances]})
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(state))
        self.ec2.describe_instances(*[i.instance_id for i in instances])
        self.mocker.result(succeed(instances))
        self.mocker.replay()

        log = self.capture_logging("ensemble.ec2")

        def validate_result(result):
            self.assertTrue(result)
            machine = result.pop()
            self.assertTrue(isinstance(machine, EC2ProviderMachine))
            self.assertEqual(machine.instance_id,
                             instances[0].instance_id)
            self.assertEquals(
                log.getvalue(),
                "Ensemble environment previously bootstrapped.\n")

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        d = bootstrap.run()
        d.addCallback(validate_result)
        return d

    def test_run_with_pending(self):
        """
        If the environment is not running, but is in a pending state,
        the underlying error from the connection attempt will pop up.
        """
        instances = [
            Instance("i-foobar", "pending", dns_name="x1.example.com")]
        state = dump({
            "zookeeper-instances":
            [i.instance_id for i in instances]})
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(state))
        self.ec2.describe_instances(*[i.instance_id for i in instances])
        self.mocker.result(succeed(instances))
        self.mocker.replay()

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        d = bootstrap.run()
        self.assertFailure(d, EnvironmentPending)
        return d

    def test_run_with_launch(self):
        """
        The provider bootstrap will launch an instance when run if there
        is no existing instance.
        """
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]

        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed(instances))
        self._mock_save()
        self.mocker.replay()

        def validate_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, EC2ProviderMachine))
            self.assertEqual(result.instance_id,
                             instances[0].instance_id)

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        d = bootstrap.run()
        d.addCallback(validate_result)
        return d

    def test_provider_bootstrap(self):
        """The ec2 provider facade instantiates and runs the bootstrap."""
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed(instances))
        self._mock_save()
        self.mocker.replay()

        def validate_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, EC2ProviderMachine))
            self.assertEqual(result.instance_id,
                             instances[0].instance_id)

        provider = self.get_provider()
        d = provider.bootstrap()
        d.addCallback(validate_result)
        return d

    def test_admin_identity_passed(self):
        """
        The admin identity information is passed to bootstrap nodes so
        agents can configure the required acl entries.
        """
        self.mocker.replay()
        config = self.get_config()
        provider = MachineProvider(self.env_name, config)
        bootstrap = EC2Bootstrap(provider)
        variables = bootstrap.get_machine_variables(
            {"machine-id": "machine-1"})
        identity = make_identity("admin:%s" % config["admin-secret"])
        self.failUnlessIn(identity, variables["scripts"][-3])
