import inspect
import os

from yaml import load

from twisted.internet.defer import succeed

from txaws.ec2.model import Instance

from ensemble.providers import ec2
from ensemble.providers.ec2.utils import (
    get_current_ami, wait_for_startup, format_cloud_ini, get_machine_options)

from ensemble.lib.testing import TestCase

SAMPLE_URI = "\
http://uec-images.ubuntu.com/query/lucid/server/released.current.txt"

SAMPLE_FILE = os.path.join(os.path.dirname(inspect.getabsfile(ec2)),
                           "tests", "data", "lucid-release.txt")


class EC2UtilsTest(TestCase):

    def test_umatched_ami(self):
        """
        If an ami is not found that matches the specifications, then
        a LookupError is raised.
        """
        page = self.mocker.replace("twisted.web.client.getPage")
        page(SAMPLE_URI)
        self.mocker.result(succeed(""))
        self.mocker.replay()
        d = get_current_ami(ubuntu_release_name="lucid")
        self.failUnlessFailure(d, LookupError)
        return d

    def test_current_ami(self):
        """The current server machine image can be retrieved."""
        page = self.mocker.replace("twisted.web.client.getPage")
        page(SAMPLE_URI)
        self.mocker.result(succeed(open(SAMPLE_FILE).read()))
        self.mocker.replay()
        d = get_current_ami(ubuntu_release_name="lucid")

        def verify_result(result):
            self.assertEqual(result, "ami-714ba518")

        d.addCallback(verify_result)
        return d

    def test_current_ami_by_region(self):
        """The current server machine image can be retrieved by region."""
        page = self.mocker.replace("twisted.web.client.getPage")
        page(SAMPLE_URI)
        self.mocker.result(succeed(open(SAMPLE_FILE).read()))
        self.mocker.replay()
        d = get_current_ami(ubuntu_release_name="lucid", region="us-west-1")

        def verify_result(result):
            self.assertEqual(result, "ami-cb97c68e")

        d.addCallback(verify_result)
        return d

    def test_current_ami_non_ebs(self):
        """
        The get_current_ami function accepts several filtering parameters
        to guide image selection.
        """
        page = self.mocker.replace("twisted.web.client.getPage")
        page(SAMPLE_URI)
        self.mocker.result(succeed(open(SAMPLE_FILE).read()))
        self.mocker.replay()
        d = get_current_ami(ubuntu_release_name="lucid", ebs=False)

        def verify_result(result):
            self.assertEqual(result, "ami-2d4aa444")

        d.addCallback(verify_result)
        return d

    def test_wait_for_instances(self):
        """
        After launching a bootstrap instance, the EC2 provider bootstrap waits
        for the instance to reach a running state.
        """
        self.log = self.capture_logging()
        instances = [
            Instance("i-foobar", "pending", dns_name="x1.example.com")]

        ec2 = self.mocker.mock()
        ec2.describe_instances(*[i.instance_id for i in instances])
        self.mocker.result(succeed(instances))

        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]

        ec2.describe_instances(*[i.instance_id for i in instances])
        self.mocker.result(succeed(instances))
        self.mocker.replay()

        d = wait_for_startup(ec2, instances, interval=0.1)

        def validate_result(result):
            self.assertEqual(result, instances)

        d.addCallback(validate_result)
        return d

    def test_format_cloud_ini_with_data(self):
        """The format cloud ini creates a user-data cloud-init config file.

        It can be used to pass an additional data dictionary as well for
        use by the launched machine.
        """
        packages = ["python-lxml"]
        scripts = ["wget http://lwn.net > /tmp/out"]
        repositories = ["ppa:ensemble/ppa"]
        output = format_cloud_ini(
            "zebra",
            packages=packages,
            scripts=scripts,
            repositories=repositories,
            data={"magic": [1, 2, 3]})

        lines = output.split("\n")
        self.assertEqual(lines.pop(0), "#cloud-config")
        config = load("\n".join(lines))
        self.assertEqual(config["ssh_authorized_keys"], ["zebra"])
        self.assertTrue(config["apt-update"])
        self.assertTrue(config["apt-upgrade"])
        formatted_repos = [dict(source=r) for r in repositories]
        self.assertEqual(config["apt_sources"], formatted_repos)
        self.assertEqual(config["runcmd"], scripts)
        self.assertEqual(config["machine-data"]["magic"], [1, 2, 3])

    def test_get_machine_options(self):
        """
        F{get_machine_options}, returns a dictionary of arguments for the ec2
        client's run instance method.
        """
        config = {}
        config["default-image-id"] = "ami-foobar"
        config["default-instance-type"] = "m1.medium"
        authorized_keys = "zebra"
        repositories = ["ppa:ensemble/ppa"]
        packages = ["python-zookeeper", "zookeeperd"]

        d = get_machine_options(
            config,
            authorized_keys=authorized_keys,
            packages=packages,
            repositories=repositories,
            magic=[1, 2, 3])

        cloud_init_config = format_cloud_ini(
            authorized_keys, packages, repositories,
            data={"magic": [1, 2, 3]})

        def verify_result(result):
            self.assertTrue(isinstance(result, dict))
            self.assertEqual(result["image_id"], "ami-foobar")
            self.assertEqual(result["user_data"], cloud_init_config)
            self.assertEqual(result["security_groups"], [])

        d.addCallback(verify_result)
        return d

    def test_get_machine_options_defaults(self):
        """
        F{get_machine_options}, returns a dictionary of arguments for
        the ec2 client's run instance method. A known set of default
        is utilized.
        """
        config = {}

        d = get_machine_options(config, "zebra")

        def verify_result(result):
            self.assertTrue(isinstance(result, dict))
            self.assertEqual(result["image_id"], "ami-d2ba45bb")
            self.assertEqual(result["instance_type"], "m1.small")

        d.addCallback(verify_result)
        return d

    def test_configured_default_image_has_precedence_over_region(self):
        """If the server machine image is specified it takes precendence over
        region."""
        d = get_machine_options(
            {"region": "eu-west-1", "default-image-id": "21down"},
            authorized_keys="")

        def verify_result(result):
            self.assertEqual(result["image_id"], "21down")

        d.addCallback(verify_result)
        return d
