import logging
import csv
import StringIO

from twisted.web.client import getPage
from twisted.internet.defer import Deferred, succeed
from yaml import safe_dump

log = logging.getLogger("ensemble.ec2")

# default image for each region.
AMI_REGION_MAP = {
    "us-east-1": "ami-d2ba45bb",
    "us-west-1": "ami-15693a50",
    "ap-northeast-1": "ami-80bb1181",
    "ap-southeast-1": "ami-b4671ee6",
    "eu-west-1": "ami-0991a67d"
    }

BASE_IMAGE_URI = "http://uec-images.ubuntu.com/query/"
CURRENT_IMAGE_URI = BASE_IMAGE_URI + \
              "%(ubuntu_release_name)s/%(variant)s/%(version)s.current.txt"


def get_region_uri(region):
    """Get the URL endpoint for the region."""
    return "https://ec2.%s.amazonaws.com" % region


def get_current_ami(ubuntu_release_name="lucid", arch="i386", ebs=True,
                    region="us-east-1", daily=False, desktop=False,
                    url_fetch=None):
    """Get the latest ami for the last release of ubuntu."""

    data = {}
    data["ubuntu_release_name"] = ubuntu_release_name
    data["version"] = daily and "daily" or "released"
    data["variant"] = desktop and "desktop" or "server"
    ebs_match = ebs and "ebs" or "instance-store"

    url = CURRENT_IMAGE_URI % data
    url_fetch = url_fetch or getPage

    d = url_fetch(url)

    def extract_ami(current_data):
        data_stream = StringIO.StringIO(current_data)
        for tokens in csv.reader(data_stream, "excel-tab"):
            if tokens[4] != ebs_match:
                continue
            if tokens[5] == arch and tokens[6] == region:
                return tokens[7]
        raise LookupError((ubuntu_release_name, arch, region,
                           data["version"], data["variant"]))

    d.addCallback(extract_ami)
    return d


def wait_for_startup(client, instances, interval=2):
    """Poll EC2 waiting for instance to transition to running state."""
    from twisted.internet import reactor

    on_started = Deferred()

    def poll_instance():
        d = client.describe_instances(*[i.instance_id for i in instances])
        d.addCallback(check_status)

    def check_status(instances):
        started = filter(lambda i: i.instance_state == "running", instances)

        if len(started) == len(instances):
            on_started.callback(instances)
        else:
            log.info("waiting for instance startup...")
            reactor.callLater(interval, poll_instance)

    reactor.callLater(interval, poll_instance)
    return on_started


def format_cloud_ini(
    authorized_keys, packages=(), repositories=None, scripts=None, data=None):
    """Format a user-data cloud-init file.

    This will enable package installation, and ssh access, and script
    execution on launch, and passing data values to an instance.

    Its important to note that no sensistive data (credentials) should be
    conveyed via cloud-config, as the values are accessible from the any
    process by default on the instance.

    Further documentation on the capabilities of cloud-init
    https://help.ubuntu.com/community/CloudInit

    @param authorized_keys The authorized SSH keys to be used when populating
           the newly launched machine's.
    @type string

    @param packages A list of packages to be installed on a machine.
    @type list of strings

    @param repository A listing of debian repostiories to add the machine's
           apt sources. 'ppa:' syntax can be utilized.
    @type list of strings

    @param scripts A list of scripts to be executed on machine start.
    @type list of strings

    @param data An optional dictionary of data to be passed via cloud config.
           It will be accessible via the key 'machine-data' from the yaml
           data structure.
    """
    cloud_config = {
        "apt-update": True,
        "apt-upgrade": True,
        "ssh_authorized_keys": [authorized_keys],
        "packages": []}

    if data:
        cloud_config["machine-data"] = data

    if packages:
        cloud_config["packages"].extend(packages)

    if repositories:
        sources = [dict(source=r) for r in repositories]
        cloud_config["apt_sources"] = sources

    if scripts:
        cloud_config["runcmd"] = scripts

    output = safe_dump(cloud_config)
    output = "#cloud-config\n%s" % (output)
    return output


def get_machine_options(config, authorized_keys, packages=None,
                        repositories=None, scripts=None, **kw):
    """Determine launch options for the machine.

    Given an ensemble environment configuration, extract the proper machine
    information to be used for launching the machine via the ec2 api.
    """
    user_data = format_cloud_ini(
        authorized_keys=authorized_keys,
        packages=packages,
        repositories=repositories,
        scripts=scripts,
        data=kw)

    instance_type = config.get("default-instance-type", "m1.small")
    # XXX ideally should come from latest available or client release name,
    # for now to speed bootstrap, we use a prepopulated image by region.
    region = config.get("region", "us-east-1")
    image_id = config.get("default-image-id", AMI_REGION_MAP[region])
    d = succeed(image_id)

    def on_machine_image(current_image_id):
        return {"image_id": current_image_id,
                "min_count": 1,
                "max_count": 1,
                "instance_type": instance_type,
                "user_data": user_data,
                "security_groups": []}
    d.addCallback(on_machine_image)
    return d
