import os

from twisted.python.failure import Failure

from ensemble.environment.tests.test_config import EnvironmentsConfigTestBase
from ensemble.errors import ProviderInteractionError, EnsembleError
from ensemble.providers.common import (
    get_user_authorized_keys, convert_unknown_error)


class CommonProviderTests(EnvironmentsConfigTestBase):

    def setUp(self):
        super(CommonProviderTests, self).setUp()
        os.mkdir(os.path.join(self.tmp_home, ".ssh"))

    def test_key_retrieval(self):
        """
        The F{ensemble.providers.common.get_user_authorized_keys} will retrieve
        a key's contents from disk and return the key data.
        """
        data = "alice in wonderland"
        self.makeFile(
            data,
            path=os.path.join(self.tmp_home, ".ssh", "identity_foobar.pub"))
        config = {"authorized-keys-path": "identity_foobar.pub"}
        key_data = get_user_authorized_keys(config)
        self.assertEqual(data, key_data)

    def test_content_utilized(self):
        """
        The entire key content can be specified via configuration.
        """
        config = {"authorized-keys": "abc"}
        key_data = get_user_authorized_keys(config)
        self.assertEqual(key_data, config["authorized-keys"])

    def test_qualified_path(self):
        """
        An alternate path can be specified with user directory expansion.
        """
        alternate_dir = os.path.join(self.tmp_home, "es")
        os.mkdir(alternate_dir)
        self.makeFile(
            "public-abc",
            path=os.path.join(alternate_dir, "ensemble_key.pub"))
        config = {"authorized-keys-path": "~/es/ensemble_key.pub"}
        key_data = get_user_authorized_keys(config)
        self.assertEqual(key_data, "public-abc")

    def test_invalid_key_specified(self):
        """If an invalid key is specified, a LookupError is raised."""
        config = {"authorized-keys-path": "zebra_moon.pub"}
        self.assertRaises(LookupError, get_user_authorized_keys, config)

    def test_convert_unknown_error(self):
        error = self.assertRaises(
            ProviderInteractionError,
            convert_unknown_error,
            OSError("Bad"))
        self.assertInstance(error, EnsembleError)
        self.assertEqual(
            str(error),
            "ProviderError: Interaction with machine provider failed: "
            "OSError('Bad',)")

    def test_convert_unknown_error_ignores_ensemble_error(self):
        error = self.assertRaises(
            EnsembleError,
            convert_unknown_error,
            EnsembleError("Magic"))
        self.assertEqual(
            str(error),
            "Magic")

    def test_convert_unknown_error_ignores_ensemble_failure(self):
        failure = convert_unknown_error(Failure(EnsembleError("Magic")))
        self.assertTrue(isinstance(failure, Failure))
        self.assertEqual(failure.value.__class__, EnsembleError)

    def test_convert_unknown_error_with_failure(self):
        failure = convert_unknown_error(Failure(OSError("Bad")))
        self.assertTrue(isinstance(failure, Failure))
        self.assertInstance(failure.value, ProviderInteractionError)
        self.assertEqual(
            str(failure.value),
            "ProviderError: Interaction with machine provider failed: "
            "OSError('Bad',)")
